with Globals, Ints, Protected_Counters;
use Globals, Ints, Protected_Counters;

pragma Elaborate_All(Globals,Ints,Protected_Counters);

package body CG.Spher is

  procedure Initialize_Regge(Parallel: in Boolean) is
  begin
    if Not_Initialized then
      if Verbosity>1 then
        Show2("Starting Initialize_ProdRegge. LMax, VSize = ",LMax,VSize);
      end if;
      Compute_Regges(Parallel);
      if Verbosity>1 then
        Show2("Done Initialize_ProdRegge.    CSize, Count = ",CSize,Count);
      end if;
      Not_Initialized := False;
    end if;
  end Initialize_Regge;

  protected body Protected_Pointer is
    procedure Check(Ind: in Integer; Ptr: out Integer) is
      P: Integer renames Pointer(Ind);
    begin
      if P<0 then          --- do work
        Count := Count+1;
        P := Count;
        Ptr := Count;      --- return positive Ptr
      else                 --- work done already
        Ptr := 0;
      end if;
    end Check;
  end Protected_Pointer;

  procedure Compute_Regges(Parallel: in Boolean) is
    --- at the end, Ptr=Pointer(Ind) is
    ---    positive if C(Prt) is nonzero
    ---    zero if C(Ptr) is zero
    ---    negatve if C(Ptr) has not been computed
    PC: Protected_Counter(0);
    PP: Protected_Pointer;

    procedure Regge_Loop is
      L1,L3,M3,Ind,Ptr: Integer;
      R: Regge;
      R11: Integer renames R(I11);
      R12: Integer renames R(I12);
      R13: Integer renames R(I13);
      R21: Integer renames R(I21);
      R22: Integer renames R(I22);
      R23: Integer renames R(I23);
      R31: Integer renames R(I31);
      R32: Integer renames R(I32);
      R33: Integer renames R(I33);
      Coe,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5: Scalar;
    begin
      Proper_Rounding;
      loop
        PC.Next(L1);
        exit when L1>Siz;
        for L2 in 0 .. Siz loop
          L3 := abs(L1-L2);
          while L3 <= IMin(LastL3,L1+L2) loop         -- abs(L1-L2) <= L3 <= IMin(LastL3,L1+L2)
            R11 := L2+L3-L1;
            R12 := L3+L1-L2;
            R13 := L1+L2-L3;
            for M1 in 0 .. L1 loop                    -- 0<=M1<=L1
              R21 := L1-M1;
              R31 := L1+M1;
              for M2 in -L2 .. L2 loop                -- |M2|<=L2
                M3 := -(M1+M2);                       -- M1+M2+M3=0
                if abs(M3)<=L3 then                   -- |M3|<=L3
                  Indx2(L1,L2,L3,M1,M2,M3,Ind);
                  if Ind>0 then                       -- work to be done
                    PP.Check(Ind,Ptr);
                    if Ptr>0 then                     -- do work, Pointer(Ind) is reserved
                      R22 := L2-M2;
                      R23 := L3-M3;
                      R32 := L2+M2;
                      R33 := L3+M3;
                      SymRegge(R,Coe,Tmp1,Tmp2,Tmp3,Tmp4,Tmp5);
                      if IsZero(Coe) then
                        Pointer(Ind) := 0;            -- done, but Coe turned out be zero
                      else
                        Copy(Coe,C(Ptr));
                      end if;
                    end if;                           -- work done
                  end if;
                end if;
              end loop; -- M2
            end loop; -- M1
            L3 := L3+2;
          end loop; -- L3
        end loop; -- L2
      end loop; -- L1
    end Regge_Loop;

    NT: constant Natural := Reserve_Tasks(Siz,Parallel);
    Task_Error: Boolean := False;
  begin
    if NT=0 then
      Regge_Loop;
    else
      declare
        task type Regge_Task_Type is end Regge_Task_Type;
        task body Regge_Task_Type is begin
          Regge_Loop;
        exception
          when others => Task_Error := True; raise;
        end Regge_Task_Type;

        Regge_Task: array(1 .. NT) of Regge_Task_Type;
        pragma Warnings (Off,Regge_Task);
      begin
        Regge_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Compute_Regges;

  procedure GetRegge(L1,L2,L3: in Natural; M1,M2,M3: in Integer; Coe: in out Scalar) is
    Ind,P: Integer;
  begin
    Indx2(L1,L2,L3,M1,M2,M3,Ind);
    if Ind=0 then                            --- violates constraints
      SetZero(Coe);
    else
      P := Pointer(Ind);
      if P>0 then
        Copy(C(P),Coe);
      elsif P=0 then
        SetZero(Coe);
      else
        raise Undefined;
      end if;
    end if;
  end GetRegge;

  procedure ProdRegge(L1,L2,L3: in Natural; M1,M2,M3: in Integer; Coe: in out Scalar) is
    Ind0,Ind,P0,P: Integer;
  begin
    Indx2(L1,L2,L3,M1,M2,M3,Ind0,Ind);
    if Ind=0 then                            --- violates constraints
      SetZero(Coe);
    else
      P0 := Pointer(Ind0);
      if P0>0 then
        P := Pointer(Ind);
        if P>0 then
          Prod(C(P0),C(P),Coe);
        elsif P=0 then
          SetZero(Coe);                      --- C^{L1,L2,L3}_{M1,M2,M3}=0
        else
          raise Undefined;
        end if;
      elsif P0=0 then
        SetZero(Coe);                        --- C^{L1,L2,L3}_{ 0, 0, 0}=0
      else
        raise Undefined;
      end if;
    end if;
  end ProdRegge;

end CG.Spher;
