with Flts, Ints, Rationals;
use Flts, Ints, Rationals;

pragma Elaborate_All (Flts,Ints,Rationals);

package Params is

   subtype String4 is String(1..4);

   type Param is
      record
         Prefix: String4;          --- name
         Map: Natural;             ---
         Alpha: Rational;
         BetaI: Integer;
         BetaJ: Positive;
         FDeg:  Natural;           --- max degree of CosSin
         Size:   Positive;         --- Size of Spheric coeff array
         Rho1:    Radius;          --- domain for TCosSin
         Rho2:    Radius;          --- domain for Spheric
         RefPhi: MinusOneZeroOne;  --- odd, no reflection, even in phi
         RefTh:  MinusOneZeroOne;  --- odd, no reflection, even wrt theta -> pi-theta
         Symm:   Integer;          --- symmetry with respect to 2pi/Symm rotations
         Mixed: Boolean;           --- False: parity Par under t --> -t, True: parity Par under (t,phi)-->(-t,-phi)
         Par: Parity;              --- see above
         FreqPar: MinusOneZeroOne; --- only odd, all, only even time frequencies
         Mat_PackCSize:  Natural;
         Contr_PackCSize: Natural;
         Mat_PackSSize:  Natural;
         Contr_PackSSize: Natural;
      end record;

   TST1: constant Param := (Prefix => "TST1",  ---- test stationary
                            Map => 0,
                            FDeg => 0,
                            Alpha => 1/1,
                            BetaI => 0,
                            BetaJ => 1,
                            Size => 20,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => 1,
                            Symm => -1,
                            Mixed => False,
                            Par => 0,
                            FreqPar => 0,
                            Mat_PackSSize => 8,
                            Contr_PackSSize => 10,
                            Mat_PackCSize => 0,
                            Contr_PackCSize => 0);

   TST2: constant Param := (Prefix => "TST2",  ----
                            Map => 0,
                            Alpha => 1/1,
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 0,
                            Size => 50,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => -1,
                            Symm => -1,
                            Mixed => False,
                            Par => 0,
                            FreqPar => 0,
                            Mat_PackSSize => 10,
                            Contr_PackSSize => 18,
                            Mat_PackCSize => 0,
                            Contr_PackCSize => 0);

   TST3: constant Param := (Prefix => "TST3",  ----
                            Map => 0,
                            Alpha => 1/1,
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 0,
                            Size => 50,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => -1,
                            Symm => -2,
                            Mixed => False,
                            Par => 0,
                            FreqPar => 0,
                            Mat_PackSSize => 28,
                            Contr_PackSSize => 30,
                            Mat_PackCSize => 0,
                            Contr_PackCSize => 0);

   PLT1: constant Param := (Prefix => "PLT1",  ----
                            Map => 1,
                            Alpha => 3/5,
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 94,
                            Size => 30,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => 1,
                            Symm => -1,
                            Mixed => False,
                            Par => 0,
                            FreqPar => -1,
                            Mat_PackSSize => 8,
                            Contr_PackSSize => 12,
                            Mat_PackCSize => 22,
                            Contr_PackCSize => 94);
   
   ROT1: constant Param := (Prefix => "ROT1", 
                            Map => 3,
                            FDeg => 0,
                            Alpha => 8/7,
                            BetaI => 0,
                            BetaJ => 1,
                            Size => 50,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => 1,
                            Symm => -3,
                            Mixed => False,
                            Par => 0,
                            FreqPar => 0,
                            Mat_PackSSize => 20,
                            Contr_PackSSize => 50,
                            Mat_PackCSize => 0,
                            Contr_PackCSize => 0);
   
   WAV1: constant Param := (Prefix => "WAV1",  ---- on modes L=3, M=1,3
                            Map => 2,
                            Alpha => 17/5, ---- 3.4, to zero when alpha -> 3.46...
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 120,
                            Size => 50,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => 1,
                            Symm => -1,
                            FreqPar => -1,
                            Mixed => False,
                            Par => 0,
                            Mat_PackSSize => 8,
                            Contr_PackSSize => 12,
                            Mat_PackCSize => 16,
                            Contr_PackCSize => 80);

   WAV2: constant Param := (Prefix => "WAV2",  ---- on modes
                            Map => 2,
                            Alpha => 17/5,   ---- 3.4, to zero when alpha -> 3.46...
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 120,
                            Size => 50,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => -1,
                            Symm => -2,
                            Mixed => False,
                            Par => 0,
                            FreqPar => -1,
                            Mat_PackSSize => 8,
                            Contr_PackSSize => 12,
                            Mat_PackCSize => 16,
                            Contr_PackCSize => 80);

   QP05: constant Param := (Prefix => "QP05",  ----  K=0.68...
                            Map => 4,
                            Alpha => 2/1,
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 250,
                            Size => 30,
                            Rho1  => Flt(33)/Flt(32),
                            Rho2  => Flt(65)/Flt(64),
                            RefPhi => 0,
                            RefTh => 1,
                            Symm => -1,
                            Mixed => True,
                            Par => 0,
                            FreqPar => -1,
                            Mat_PackSSize => 8,
                            Contr_PackSSize => 20,
                            Mat_PackCSize => 16,
                            Contr_PackCSize => 140);

   Q1138: constant Param := (Prefix => "1138",  ---- K=0.842
                             Map => 4,
                             Alpha => 1/1,
                             BetaI => -1,
                             BetaJ => 2,
                             FDeg => 400,
                             Size => 30,
			     Rho1  => Flt(33)/Flt(32),
			     Rho2  => Flt(65)/Flt(64),
                             RefPhi => 0,
                             RefTh => 0,
                             Symm => -1,
                             Mixed => True,
                             Par => 0,
                             FreqPar => -1,
                             Mat_PackSSize => 12,
                             Contr_PackSSize => 16,
                             Mat_PackCSize => 64,
                             Contr_PackCSize => 350);

   Q2502: constant Param := (Prefix => "2502",  ---- K=0.8
                             Map => 4,
                             Alpha => 2/1,
                             BetaI => 1,
                             BetaJ => 2,
                             FDeg => 280,
                             Size => 36,
			     Rho1  => Flt(33)/Flt(32),
			     Rho2  => Flt(65)/Flt(64),
                             RefPhi => 0,
                             RefTh => 1,
                             Symm => -1,
                             Mixed => True,
                             Par => 0,
                             FreqPar => -1,
                             Mat_PackSSize => 12,
                             Contr_PackSSize => 16,
                             Mat_PackCSize => 48,
                             Contr_PackCSize => 256);

   Q3601: constant Param := (Prefix => "3601",  ---- K=0.53
                             Map => 4,
                             Alpha => 3/2,
                             BetaI => 1,
                             BetaJ => 2,
                             FDeg => 300,
                             Size => 40,
			     Rho1  => Flt(33)/Flt(32),
			     Rho2  => Flt(65)/Flt(64),
                             RefPhi => 0,
                             RefTh => 1,
                             Symm => -1,
                             Mixed => True,
                             Par => 0,
                             FreqPar => -1,
                             Mat_PackSSize => 10,
                             Contr_PackSSize => 16,
                             Mat_PackCSize => 36,
                             Contr_PackCSize => 240);

   ALLX: constant Param := (Prefix => "ALLX",  -----  dummy
                            Map => 0,
                            Alpha => 1/1,
                            BetaI => 0,
                            BetaJ => 1,
                            FDeg => 10,
                            Size => 10,
			    Rho1  => Flt(33)/Flt(32),
			    Rho2  => Flt(65)/Flt(64),
                            RefPhi => 1,
                            RefTh => 1,
                            Symm => 1,
                            Mixed => True,
                            Par => 0,
                            FreqPar => -1,
                            Mat_PackSSize => 12,
                            Contr_PackSSize => 16,
                            Mat_PackCSize => 0,
                            Contr_PackCSize => 0);

   type Param_List is array(Natural range <>) of Param;

   PList: constant Param_List(0..11) := (ALLX,TST1,TST2,TST3,PLT1,ROT1,WAV1,WAV2,
					 QP05,Q1138,Q2502,Q3601);

   procedure Set_Defaults(P: out Param);
   procedure Show(N: in String; P: in Param);
   function Left_Part(S: String; L: Positive) return String;
   function IndexOf(S: String) return Natural;

end Params;
