with Ints, Make_Count;
use Ints;

pragma Elaborate_All (Ints,Make_Count);

generic

  LMax: in Positive;

package Regges is

  I11: constant Integer := 0;
  I12: constant Integer := 1;
  I13: constant Integer := 2;
  I21: constant Integer := 3;
  I22: constant Integer := 4;
  I23: constant Integer := 5;
  I31: constant Integer := 6;
  I32: constant Integer := 7;
  I33: constant Integer := 8;

  subtype Regge is Int_Vec(I11 .. I33); -- represents a 3x3 matrix

  --- N1,N2,N3,M1,M2,M3 are doubled 3j arguments
  procedure To3j(R: in Regge; N1,N2,N3: out Natural; M1,M2,M3: out Integer);
  procedure ToRegge(N1,N2,N3: in Natural; M1,M2,M3: in Integer; R: out Regge; Is_Zero: out Boolean);
  function Check(R: Regge) return Boolean;
  procedure Order(R: in out Regge; ChangeSign: out Boolean);

  function Indx(S,B,T,X,L: Natural) return Positive;
  procedure CheckIndx;
  procedure Indx(R: in Regge; I: out Positive; ChangeSign: out Boolean);
  procedure Indx(N1,N2,N3: in Natural; M1,M2,M3: in Integer; I: out Natural; ChangeSign: out Boolean);
  function Indx(N1,N2,N3: Natural; M1,M2,M3: Integer) return Natural;
  procedure OrderedRegge(S,B,T,X,L: in Natural; R: out Regge);

  procedure Show(R: in Regge);
  procedure Show(N1,N2,N3: in Natural; M1,M2,M3: in Integer);
  procedure ShowBoth(R: in Regge);

  --- L1,L2,L3,M1,M2,M3 are true 3j arguments (integers)
  procedure Ordered(L1,L2,L3: in Natural; S,B,T,X,L: out Natural; Is_Zero: out Boolean);
  procedure Ordered(L1,L2,L3: in Natural; M1,M2,M3: in Integer; S,B,T,X,L: out Integer; Is_Zero: out Boolean);
  procedure Indx2(L1,L2,L3: in Natural; I0: out Natural);
  procedure Indx2(L1,L2,L3: in Natural; M1,M2,M3: in Integer; I: out Natural);
  procedure Indx2(L1,L2,L3: in Natural; M1,M2,M3: in Integer; I0,I: out Natural);

  --- for testing
  procedure ShowLM(N: in String; L1,L2,L3,M1,M2,M3: in Integer);
  procedure CheckLM(N: String; L1,L2,L3,M1,M2,M3: Integer);

  LMax1:    constant Positive := LMax+1;
  LMax2:    constant Positive := LMax1*LMax1;
  LST_Count: constant Int_Vec := Make_Count(LMax);
  VSize:    constant Positive := LST_Count(LST_Count'Last);

  Inconsistent: exception;

  pragma Inline_Always (To3j,Check,OrderedRegge);

end Regges;
