with ScalVectors, Linear, Linear.Contr;

pragma Elaborate_All (ScalVectors,Linear.Contr);

generic

  with procedure Put_Numeric(N: in Positive; S: in Scalar) is <>;
  with procedure Get_Numeric(N: in Positive; S: in out Scalar) is <>;

package TSpherics.Fix is

  procedure LinMap(LF: in out Spheric);
  procedure GMap(F: in Spheric; GF: in out Spheric);
  procedure DGMap(F,FF3,H: in Spheric; DGH: in out Spheric);
  procedure DContrNorm(M: in SMatrix; F: in Spheric; PackCSize,PackSSize: in Natural; R: out Radius);
  procedure ContrFix(M: in SMatrix; PackCSize,PackSSize: in Natural; F: in out Spheric);

  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SV_Ops renames SSV.SV_Ops;
  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => Spheric,
                             Mode => TSMode, Modes => TSModes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);

  --- numerical
  procedure DerMatrix(F: in Spheric; M: in out SMatrix);
  procedure ContrMatrix(F: in Spheric; M: in out SMatrix);
  procedure ApproxFix(F: in out Spheric; M: in out SMatrix; Eps: in Radius; MinErr: in out Radius);
  procedure ApproxFix(F: in out Spheric; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(F: in out Spheric; M: in out SMatrix);
  procedure EigenDer(Name: in String; F: in Spheric);

  QC: TSModes := Make(Mat_PackCSize,Mat_PackSSize,True);   --- no Err modes
  subtype QC_Vector is SV.Vector(QC'Range);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

end TSpherics.Fix;
