with Ada.Text_IO, Logicals, ScalInfo, Ints, Rationals, Flts;
use Ada.Text_IO, Logicals, ScalInfo, Ints, Rationals, Flts;
with Params.Names, Matrices, Vectors, Polynoms, TCosSins1, CG, CG.Spher, TSpherics_Aux;
use Params;

pragma Elaborate_All (Ada.Text_IO,Logicals,ScalInfo,Ints,Rationals,Flts);
pragma Elaborate_All (Matrices,Vectors,Polynoms,TCosSins1,CG.Spher,TSpherics_Aux);

generic  --- instantiate CosSin1 --> SCoeff

  Trunc: in Boolean;            --- True --> Numeric, False --> Ball
  Full: in Boolean := False;    --- all spatial modes
  Pram: in Param;

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure IPower(I: in Integer; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;
  --with function Pi return Scalar is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  with function Get_Precision(R: Scalar) return Positive is <>;

  -------------------- end_include scalars

  with procedure Proper_Rounding is <>;

package TSpherics is

  pragma Elaborate_Body;

  package PN is new Params.Names(P => Pram, Trunc => Trunc);
  use PN;

  procedure Show_Param(All_Param: in Boolean := True);

  package SV is new Vectors (Component => Scalar);
  package SM is new Matrices (Scalar => Scalar);
  subtype SVector is SV.Vector;
  subtype SMatrix is SM.Matrix;
  subtype CRange  is Integer range 0 .. Size;
  subtype SArray is SMatrix(CRange,CRange);

  --- values of scaled associated Legendre functions P^m_l with 0 <= m <= l
  procedure ALegend_Vals(C: in Scalar; V: in out SArray);    -- V(l,m)=P^m_l(c) if s=Sqrt(1-c*c)
  procedure ALegendVals(Theta: in Scalar; V: in out SArray); -- the above with c=cos(theta)

  package Poly is new Polynoms (Scalar => Scalar, SV => SV);
  package SCoeffs is new TCosSins1(Trunc => Trunc, FDeg => FDeg, Rad => Rho1, Scalar => Scalar, SV => SV, Poly => Poly);
  subtype SCoeff is SCoeffs.CosSin1;
  subtype SCoeff2 is SCoeffs.Fourier1;
  type CVector is array(Integer range <>) of SCoeff;
  type CMatrix is array(Integer range <>, Integer range <>) of SCoeff;
  type CMatrix2 is array(Integer range <>, Integer range <>) of SCoeff2;
  subtype CArray is CMatrix(CRange, CRange);

  SizeE: constant Positive := 2*Size;
  subtype ERange is Integer range 0 .. SizeE;
  subtype EArray is RadVec(ERange);

  type Spheric is
    record
      C: CArray; --- coefficients, see below
      E: EArray; --- error norms, E(n) includes weights l>=n
    end record;

  --- our complex spherical harmonics are Y^m_l(cos(theta))=P^m_l(theta)exp(im*phi) |m|\le l
  --- for real-valued F we only need m>=0
  ---
  --- F(theta,phi) = Sum_{l,m<=l} C(l,l-m) P^m_l(theta)cos(m*phi)
  ---              + Sum_{l,m <l} C(l-m,l) P^m_l(theta)sin(m*phi)
  ---
  ---       1 coeff is at  (l,l)    on  diagonal (index2 = index1)
  --- cos(m.) coeff is at (l,l-m) below diagonal (index2 < index1)
  --- sin(m.) coeff is at (l-m,l) above diagonal (index2 > index1)

  function "="(F1,F2: Spheric) return Boolean;
  procedure FillZero(F: in out Spheric); --- keeps parities
  procedure SetZero(F: in out Spheric);  --- uses Par and Mixed as defaults
  procedure SetZero(P: in Parity; Mix: in Boolean; F: in out Spheric);
  procedure SetParity(P: in Parity; Mix: in Boolean; F: in out Spheric);
  function IsZero(F: Spheric) return Boolean;
  function IsSharp(F: Spheric) return Boolean;
  procedure Copy(F1: in Spheric; F2: in out Spheric);
  procedure Swap(F1,F2: in out Spheric);

  procedure Neg(F: in out Spheric);
  procedure Neg(F1: in Spheric; F2: in out Spheric);
  function "-"(F: Spheric) return Spheric;
  procedure Add(S: in SCoeff; F: in out Spheric);
  procedure Add(F1: in Spheric; F2: in out Spheric);
  procedure Sum(F1,F2: in Spheric; F3: in out Spheric);
  function "+"(F1,F2: Spheric) return Spheric;
  procedure Sub(F1: in Spheric; F2: in out Spheric);
  procedure Diff(F1,F2: in Spheric; F3: in out Spheric);
  function "-"(F1,F2: Spheric) return Spheric;
  procedure Mult(S: in Scalar; F: in out Spheric);
  procedure Mult(S: in SCoeff; F: in out Spheric);
  procedure Prod(S: in Scalar; F1: in Spheric; F2: in out Spheric);
  procedure Prod(S: in SCoeff; F1: in Spheric; F2: in out Spheric);
  function "*"(S: Scalar; F: Spheric) return Spheric;
  function "*"(S: SCoeff; F: Spheric) return Spheric;
  procedure AddProd(S: in Scalar; F1: in Spheric; F2: in out Spheric);
  procedure AddProd(S: in SCoeff; F1: in Spheric; F2: in out Spheric);
  procedure Mult(F1: in Spheric; F2: in out Spheric);
  procedure Prod(F1,F2: in Spheric; P3: in Parity; F3: in out Spheric); --- when parity is known
  procedure Prod(F1,F2: in Spheric; F3: in out Spheric);
  function "*"(F1,F2: Spheric) return Spheric;

  function MaxNorm(F: Spheric) return Radius;
  procedure Norm(F: in Spheric; S: in out Scalar);
  function Norm(F: Spheric) return Scalar;
  procedure Coeff(L: in Natural; M0: in Integer; F: in Spheric; S: in out SCoeff);
  procedure AddCoeff(L: in Natural; M0: in Integer; S: in SCoeff; F: in out Spheric);
  procedure AddHigher(L: in Natural; E: in Radius; F: in out Spheric);
  procedure AddBall(R: in Radius; F: in out Spheric);
  procedure AddBall(S: in Scalar; F: in out Spheric);
  procedure Show_Parities(N: in String; F: in Spheric; Hide0: in Boolean);
  procedure Show_Guessed_Radii(N: in String; F: in Spheric; Cut: in Radius);
  procedure Show(N: in String; F: in Spheric; Cut: in Radius; Hide0: in Boolean := True);
  procedure Show1(N: in String; F: in Spheric; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; H: in Spheric; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; H: in out Spheric; Decimal: in Boolean := False);
  procedure Write(FileName: in String; H: in Spheric; Decimal: in Boolean := False);
  procedure Read(FileName: in String; H: in out Spheric; Decimal: in Boolean := False);

  procedure Leading_Coeffs(Str: in String; P: in SCoeff; Cut: in Radius); --- aux
  procedure Leading_Coeffs(F: in Spheric; Cut: in Radius);                --- info

  procedure Val(F: in Spheric; Theta,Phi: in Scalar; S: in out SCoeff);
  procedure Val(F: in Spheric; T,Theta,Phi: in Scalar; S: in out Scalar);
  procedure Val2(F: in Spheric; Theta,Phi: in Scalar; S: in out SCoeff2);
  procedure Val2(F: in Spheric; T,Theta,Phi: in Scalar; S: in out Scalar);
  procedure NegInvLap(F: in out Spheric);
  procedure NegInvLap(F1: in Spheric; F2: in out Spheric);

  --- periodic solutions
  procedure InvPlate(F: in out Spheric);
  procedure InvWave(F: in out Spheric);
  procedure RotWave(F: in out Spheric);  --- for FDeg=0, rational Alpha

  --- quasiperiodic
  procedure InvQP(F: in out Spheric);

  --- symmetries
  procedure FlipNorthSouth(F: in out Spheric);
  procedure Reflect0(F: in out Spheric);
  procedure PhiRotatePi(F: in out Spheric);
  procedure ProjReflect0(Ref: in MinusOneZeroOne; F: in out Spheric);
  procedure ProjPhi(Ref: in MinusOneZeroOne; F: in out Spheric);
  procedure ProjNorthSouth(Ref: in MinusOneZeroOne; F: in out Spheric);
  procedure ProjTheta(Ref: in MinusOneZeroOne; F: in out Spheric) renames ProjNorthSouth;
  procedure ProjMod(Base: in Positive; Dig: in Natural; F: in out Spheric);
  procedure ProjM0(F: in out Spheric);
  procedure ProjSymm(Rot: in Integer; F: in out Spheric);
  procedure ProjFreq(P: in MinusOneZeroOne; F: in out Spheric);
  procedure ProjParity(P: in Parity; Mix: in Boolean; F: in out Spheric; Check: in Boolean := True);
  procedure ProjStat(F: in out Spheric);
  procedure ProjAll(F: in out Spheric);

  procedure Parities(F: in Spheric; CosEven,CosOdd,SinEven,SinOdd: out Boolean);
  procedure CheckParities(F: in Spheric; Mix: in Logical := Uncertain);
  procedure Sanitize_Parities(F: in out Spheric);
  procedure ForceParity(P: in Parity; Mix: in Boolean; F: in out Spheric) renames SetParity;
  procedure Guess_Parity(F: in Spheric; Mix: in Boolean; P: out Parity; Is_Zero: out Boolean);

  --- purely numerical
  procedure Simple_Random(F: in out Spheric; P: in Parity; Mix: in Boolean; Fac: in Radius := Two);
  procedure NumTrunc(SizC,SizS: in Natural; F: in out Spheric);
  procedure NumTrunc(SizC,SizS: in Natural; F1: in Spheric; F2: in out Spheric);

  --- plotting
  procedure Make_PlotData(F: in Spheric; P: in out CMatrix2);                   --- SCoeff2 values on (Theta,Phi) grid
  procedure SinProjection(Theta,Phi: in Flt; X,Y: out Flt);                     --- sinusoidal projection
  procedure Plot(FileName: in String; P: in CMatrix2; T: in Flt);               --- plot single time T
  procedure Plot(FileName: in String; F: in Spheric; T: in Flt;
                  ThetaSteps: in Positive := 128; PhiSteps: in Positive := 64); --- plot single time T, do not use in loops
  procedure Plot(Stem: in String; F: in Spheric; ThetaSteps,PhiSteps,TSteps: in Positive); --- plot TStep times/files

  procedure Plot2(Stem: in String; F: in Spheric; ThetaSteps,PhiSteps,TSteps: in Positive) renames Plot;

  --- modes
  type TSMode is      --- a unit coeff or unit error
    record
      N: Natural;     --- CosSins1 index (EFreq may be empty)
      I: ERange;      --- 1st Spheric index
      J: ERange;      --- 2nd Spheric index; notice L=IMax(I,J), K=IMin(I,J), M=L-K
      CErr: Boolean;  --- CosSins1 error
      SErr: Boolean;  --- Spheric error
    end record;

  type TSModes is array (Positive range <>) of TSMode; -- should start with index 1

  procedure Norm(A: in TSMode; S: in out Scalar);
  procedure MaxNorm(A: in TSMode; S: out Radius);
  procedure Assign(A: in TSMode; F: in out Spheric);
  procedure AddProd(A: in TSMode; S: in Scalar; F: in out Spheric; Check: in Boolean);
  procedure Extract_Coeff(A: in TSMode; F: in out Spheric; S: in out Scalar);
  procedure Extract_Err(A: in TSMode; F: in out Spheric; S: in out Scalar);
  procedure Extract(A: in TSMode; F: in out Spheric; S: in out Scalar; Check: in Boolean);
  procedure Show(A: in TSMode; NewLine: in Boolean := False);
  function Symm_Check(A: TSMode) return Boolean;
  function Make(CSize: Natural; SSize: Positive; No_Errs: Boolean) return TSModes;

  --- Clebsch-Gordan and Regge symbols
  LastL3: constant Integer := Choose(Trunc,Size,SizeE);
  LMax: constant Positive := 2*LastL3;                     --- largest Regge entry
  package NewCG is new CG (LMax => LMax, Scalar => Scalar);
  package New_Regges renames NewCG.CGR;
  package NewCG_Spher is new NewCG.Spher (Siz => Size);

  --- various bounds, also SAlpha, SAlpha2, SInvAlpha
  package TSp_Aux is new TSpherics_Aux (Trunc => Trunc, Alpha => Alpha, Scalar => Scalar);
  use TSp_Aux;

  SRho:  constant Scalar := Scal(Rho2);
  SBeta: constant Scalar := Make_Beta(BetaI,BetaJ);

  LastRho2Power: constant Natural := 2*SizeE;  --- check if that much is needed
  InvRho2: constant Radius := One/Rho2;
  Rho2Power: constant RadVec := PowersOf(Rho2,LastRho2Power);
  InvRho2Power: constant RadVec := PowersOf(InvRho2,LastRho2Power);

  Inconsistent_Modes: exception;
  PProd_Parallel: Boolean := True;
  Check_Parities: Boolean := False;

  use SCoeffs, NewCG_Spher;

private

  Not_Trunc: constant Boolean  := not Trunc;
  SOne:      constant Scalar   := Scal(One);
  SHalf:     constant Scalar   := Scal(Half);
  SThree:    constant Scalar   := Scal(Three);

  ------------------- associated Legendre functions
  procedure ALegend_Vals(C,S: in Scalar; V: in out SArray); --- format is V(L,M)

  function IsZero(C: CArray) return Boolean;
  function IsZero(E: EArray) return Boolean;
  function EffLast(C: CArray) return Integer;
  function EffLast(E: EArray) return Integer;
  function MaxNorm(C: CArray) return Radius;
  function Norm(E: EArray) return Radius;

  type FVector is array(Integer range <>) of Flt;
  type FMatrix is array(Integer range <>, Integer range <>) of Flt;
  LastB:  constant Integer := Choose(Trunc,-1,SizeE);
  subtype BRange  is Integer range 0 .. LastB;
  subtype FArray is FMatrix(CRange, CRange);
  subtype BigEArray is FMatrix(BRange, BRange);

  procedure MaxNorms(C: in CArray; A: out FArray);
  procedure AddNorms(B: in BigEArray; E: in out EArray);
  procedure Double(M3: in ERange; C3: in out CArray; B3: in out BigEArray);

  procedure AddProdC(C1,C2: in CArray; A1,A2: in FArray; F3: in out Spheric);
  procedure AddProdC(M3: in ERange; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Sum(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Diff(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Sum_CosCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Diff_CosCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Sum_CosSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Diff_CosSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Sum_SinCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Diff_SinCos(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Sum_SinSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProd_Diff_SinSin(M1,M2,M3: in Natural; C1,C2: in CArray; A1,A2: in FArray; C3: in out CArray; B3: in out BigEArray);
  procedure AddProdE(A1: in FArray; E2: in EArray; E3: in out EArray);

  --- periodic
  procedure InvWave(F: in out Spheric; WDel: in EArray);
  procedure RotWave(F: in out Spheric; WDel: in EArray);
  Wave_Delta: EArray;

  --- quasiperiodic
  type CCErrs is array(CRange,CRange) of Errs1;
  procedure QP_InvHelm(L,M: in Positive; P: in Parity; QSig,QDel: in Errs1; C,S: in out SCoeff);
  procedure InvQP(F: in out Spheric; QSigDel: in CCErrs; QLam: in EArray);
  QP_SigmaDelta: CCErrs;
  QP_Lambda: EArray;

  --- plotting
  PiHalf: constant := 1.57079_63267_94896_61923_13216_91639_75144;
  PiFull: constant Flt := Two*PiHalf;
  TwoPi: constant Flt := Four*PiHalf;

end TSpherics;
